from __future__ import division

from operator import getitem

import dask.array as da
import dask.dataframe as dd
import numpy as np
from sklearn.preprocessing import label as sklabel
from sklearn.utils.validation import check_is_fitted


class LabelEncoder(sklabel.LabelEncoder):

    __doc__ = sklabel.LabelEncoder.__doc__

    def _check_array(self, y):
        if isinstance(y, dd.Series):
            y = da.asarray(y)
        return y

    def fit(self, y):
        y = self._check_array(y)

        if isinstance(y, da.Array):
            classes_ = da.unique(y)
            classes_ = classes_.compute()
        else:
            classes_ = np.unique(y)

        self.classes_ = classes_

        return self

    def fit_transform(self, y):
        return self.fit(y).transform(y)

    def transform(self, y):
        check_is_fitted(self, 'classes_')
        y = self._check_array(y)

        if isinstance(y, da.Array):
            return da.map_blocks(np.searchsorted, self.classes_, y,
                                 dtype=self.classes_.dtype)
        else:
            return np.searchsorted(self.classes_, y)

    def inverse_transform(self, y):
        check_is_fitted(self, 'classes_')
        y = self._check_array(y)

        if isinstance(y, da.Array):
            return da.map_blocks(getitem, self.classes_, y,
                                 dtype=self.classes_.dtype)
        else:
            y = np.asarray(y)
            return self.classes_[y]



import copy
from distutils.version import LooseVersion

import dask
import dask.array as da
from dask.base import tokenize
from dask.delayed import delayed, Delayed

from sklearn.utils.validation import indexable, _is_arraylike


if LooseVersion(dask.__version__) > '0.15.4':
    from dask.base import is_dask_collection
else:
    from dask.base import Base

    def is_dask_collection(x):
        return isinstance(x, Base)


def _indexable(x):
    return indexable(x)[0]


def _maybe_indexable(x):
    return indexable(x)[0] if _is_arraylike(x) else x


def to_indexable(*args, **kwargs):
    """Ensure that all args are an indexable type.

    Conversion runs lazily for dask objects, immediately otherwise.

    Parameters
    ----------
    args : array_like or scalar
    allow_scalars : bool, optional
        Whether to allow scalars in args. Default is False.
    """
    if kwargs.get('allow_scalars', False):
        indexable = _maybe_indexable
    else:
        indexable = _indexable
    for x in args:
        if x is None or isinstance(x, da.Array):
            yield x
        elif is_dask_collection(x):
            yield delayed(indexable, pure=True)(x)
        else:
            yield indexable(x)


def to_keys(dsk, *args):
    for x in args:
        if x is None:
            yield None
        elif isinstance(x, da.Array):
            x = delayed(x)
            dsk.update(x.dask)
            yield x.key
        elif isinstance(x, Delayed):
            dsk.update(x.dask)
            yield x.key
        else:
            assert not is_dask_collection(x)
            key = 'array-' + tokenize(x)
            dsk[key] = x
            yield key


def copy_estimator(est):
    # Semantically, we'd like to use `sklearn.clone` here instead. However,
    # `sklearn.clone` isn't threadsafe, so we don't want to call it in
    # tasks.  Since `est` is guaranteed to not be a fit estimator, we can
    # use `copy.deepcopy` here without fear of copying large data.
    return copy.deepcopy(est)


def unzip(itbl, n):
    return zip(*itbl) if itbl else [()] * n



import contextlib

import packaging.version
import sklearn
import dask


SK_VERSION = packaging.version.parse(sklearn.__version__)
HAS_MULTIPLE_METRICS = SK_VERSION >= packaging.version.parse('0.19.0')

DASK_VERSION = packaging.version.parse(dask.__version__)


@contextlib.contextmanager
def dummy_context(*args, **kwargs):
    yield



"""Utilities for hyperparameter optimization.

These estimators will operate in parallel. Their scalability depends
on the underlying estimators being used.
"""
from ._search import (
    GridSearchCV, RandomizedSearchCV,
    compute_n_splits, check_cv
)
from ._split import ShuffleSplit, train_test_split


__all__ = [
    'GridSearchCV',
    'RandomizedSearchCV',
    'ShuffleSplit',
    'train_test_split',
    'compute_n_splits',
    'check_cv',
]
